/*  @(#)mpoll.c	1.5  97/12/23  */

/*
 * Copyright (c) Communications Research Centre, Government of Canada 1997. 
 * All rights reserved.
 * 
 * License is granted to copy and use this software, for research and
 * evaluation purposes, provided that the Communications Research Centre
 * is acknowledged in all documentation pertaining to any such copy or
 * use. The Communications Research Centre grants no other licenses
 * expressed or implied. Title, ownership rights, and intellectual
 * property rights in the software shall remain with the Communications
 * Research Centre.  The Centre's name and the Government of Canada
 * should not be used in any advertising without written permission.
 * 
 * THE COMMUNICATIONS RESEARCH CENTRE MAKES NO REPRESENTATIONS CONCERNING
 * EITHER THE MERCHANTABILITY OF THIS SOFTWARE OR THE SUITABILITY OF
 * THIS SOFTWARE FOR ANY PARTICULAR PURPOSE.  The software is provided
 * "as is" without express or implied warranty of any kind.
 * 
 * These notices must be retained in any copies of any part of this
 * software.
 */


#include "mpoll.h"
#ifndef WIN32
#include <sys/utsname.h>
#endif

#ifdef WIN32
extern int WinPutsCmd(ClientData, Tcl_Interp*, int, char **);
#endif

/* the main data structures for topics, responses, and
   also my local responses */
    
struct topic topiclist[MAX_TOPICS+1]; 
struct response responselist[MAX_TOPICS+1*MAX_RESPONSES*1000]; 
struct response myresponse[MAX_TOPICS+1*MAX_RESPONSES+1];

/* counters for the topics and responses */
int numtopics = 0, topicnum = 0;
int numresponses = 0, responsenum = 0;
int nummyresponses = 0, myresponsenum = 0;

/* handle topics that are deleted */
int numdeleted;
struct topic deltopics[MAX_TOPICS+1];

/* counter for number of members */
int nummembers;

/* and structure for member information */
struct member memberlist[MAX_MEMBERS+1];

 
/* SCCS version information */
char vernum[] = "1.5";

/* scratch space */
char messbuf[1024];

/* info about local host and user */
char hostname[TMPSTRLEN];
char hostip[TMPSTRLEN];
/* local address number in network byte order */
unsigned long hostaddr; 
char username[TMPSTRLEN];
char systeminfo[TMPSTRLEN];

/* command line variables */
char *address = NULL;
#ifdef WIN32
u_short port;
int ttl;
#else
int port;
char ttl;
#endif

int logging = 0; /* flag for logging debugging information */

/* flag for simulating data for demos */
int simulate = 0;

/* variables to share with TCL */
char *theme;
char *version;
char *current_summary;
char *current_question;
char *current_author;
char *current_option;
char *current_time;
int current_decay;
char *default_font;
char *small_font;

/* -------------------------------------------------------------- */
#ifdef WIN32
void Et_Main(int argc, char **argv)
#else
void main(int argc, char **argv)
#endif
   {
   int arg, use_defaults;
   char *address_end;
   
   /* setup the random number generator */
   srand( (unsigned)time( NULL ) );

#ifdef WIN32
   init_win_sockets();  /* start the WIN32 socket library */
#endif
   
   learn_local_info(); /* hostname and user info */

   /* startup TCL/Tk processing */
#ifndef WIN32   
   Et_Init(&argc,argv);
#endif
   
   /* learn the new TCL commands in this file, and other C files */
   ET_INSTALL_COMMANDS;
   ET_INSTALL_COMMANDS( results.c );
   ET_INSTALL_COMMANDS( save.c );
   ET_INSTALL_COMMANDS( submit.c );
   ET_INSTALL_COMMANDS( view.c );
   
   #ifdef WIN32
   /* build a new puts command for WIN32 since we don't have a stdout */
   Tcl_CreateCommand(Et_Interp, "puts", WinPutsCmd, 0, 0);
   #endif
   
   /* load in the Tcl help procs */
   ET_INCLUDE( help.tcl );
   
   init_bitmaps(); /* setup bitmaps used by TCL */
  
   /* setup the shared C/TCL variables */
   /* do it while the char pointers are NULL */
   
   Tcl_LinkVar(Et_Interp, "theme", (char *)&theme, TCL_LINK_STRING);
   theme = strdup("Unknown Theme"); /* a default */
   
   Tcl_LinkVar(Et_Interp, "version", (char *)&version, TCL_LINK_STRING);
   version = (char *)malloc(256);
   sprintf(version, "Beta %s", vernum);
   
   Tcl_LinkVar(Et_Interp, "summary", (char *)&current_summary, TCL_LINK_STRING);
   current_summary = (char *)malloc(TMPSTRLEN);
   
   Tcl_LinkVar(Et_Interp, "question", (char *)&current_question, TCL_LINK_STRING);
   current_question = (char *)malloc(QUESTION_SIZE);
   
   Tcl_LinkVar(Et_Interp, "author", (char *)&current_author, TCL_LINK_STRING);
   current_author = (char *)malloc(TMPSTRLEN);
   
   Tcl_LinkVar(Et_Interp, "create_time", (char *)&current_time, TCL_LINK_STRING);
   current_time = (char *)malloc(TMPSTRLEN);
   
   Tcl_LinkVar(Et_Interp, "option", (char *)&current_option, TCL_LINK_STRING);
   current_option = (char *)malloc(TMPSTRLEN);
   
   Tcl_LinkVar(Et_Interp, "decaytime", (char *)&current_decay, TCL_LINK_INT);
   
   Tcl_LinkVar(Et_Interp, "default_font", (char *)&default_font, TCL_LINK_STRING);
   default_font = strdup(DEFAULT_FONT);

   Tcl_LinkVar(Et_Interp, "small_font", (char *)&small_font, TCL_LINK_STRING);
   small_font = strdup(SMALL_FONT);

   use_defaults = USEDEFAULTS;  
   /* now handle non-ET command line arguments */
   if (argc > 1)
      {
      for (arg = 1; arg < argc; ++arg)
         {
         if ( !strcmp("-C", argv[arg]))
            {
            theme = strdup(argv[arg+1]);
            ++arg;
            }
         else if (!strcmp("-t", argv[arg]))
            {
            ttl = atoi(argv[arg+1]);
            ++arg;
            }
         else if (!strcmp("-D", argv[arg]))
            {
            use_defaults = 1;
            }
         else if (!strcmp("-V", argv[arg]))
            {
            printf("%s\n", version);
            exit(1);
            }
         else if (!strcmp("-l", argv[arg]))
            {
            logging = 1;
            }
         else if (!strcmp("-s", argv[arg]))
            {
            simulate = 1;
            }
         else if (!strncmp("2", argv[arg], 1))
            {
            address = strdup(argv[arg]);
            
            if (strchr(address, '/') == NULL)
               {
               usage("Multicast address and port needed.");
               }
            
            /* terminate the address string */
            address_end = address;
            while (*address_end != '\0' && *address_end != '/')
               {
               ++address_end;
               }
            *address_end = '\0';
            port = atoi(address_end + 1);
            }
         else
            {
            sprintf(messbuf, "Unknown command line argument: %s\n", argv[arg]);
            usage(messbuf);
            }
         }
      }
      
   /* open logging window */
   if (logging)
      viewlog();
   
   
   /* create the sockets */
   if (! ttl)
      {
      ttl = DEFAULT_TTL;
      }
   if (use_defaults)
      {
      if (address == NULL)
         {
         address = strdup(DEFAULT_ADDRESS);
         }
      if (! port)
         {
         port = DEFAULT_PORT;
         }
      }
   
   if ((address && port && ttl) == 0)
      usage("Command parameter missing");
         
   init_socket();
   
   /* setup temporary parsing variables */
   
   setparsevars();
   
   /* read in the data from cache file*/
   readdata(); 
   
   /* the main TCL topics window is created here */ 
   ET_INCLUDE( topics.tcl );
   
   /* and then we populate it with the list of topics */
   listtopics();
   
   /* prepare to interrupt when something on the receive socket */

#ifdef WIN32
   Tcl_CreateFileHandler(Tcl_GetFile((ClientData)recvsock, TCL_WIN_SOCKET), TK_READABLE, (Tcl_FileProc*)hear, (ClientData)recvsock);
#elif IRIX
   Tcl_CreateFileHandler(recvsock, TK_READABLE, (Tcl_FileProc*)hear, (ClientData)recvsock);
#else
   Tcl_CreateFileHandler(Tcl_GetFile((ClientData)recvsock, TCL_UNIX_FD), TK_READABLE, (Tcl_FileProc*)hear, (ClientData)recvsock);
#endif

   /* queue my questions in the cache for retransmission */
   queue_all_questions();
   
   /* get set for retransmission of responses*/
   queue_all_responses();
   
   /* start sending out the membership packets */
   queue_member();
   
   /* This is the song that never ends... */
#ifndef WIN32   
   Et_MainLoop();
#endif
   
   }

/* -------------------------------------------------------------- */
/* show the list of topics */
listtopics()
   {
   int i;
   
   
   /* clear the list of topics */
   ET(.listframe.list delete 0 end);
   ET(update idletasks);
   
   /* build the list */
   
   for(i = 1; i <= numtopics; i++)
      {
      ET(.listframe.list insert end "%s(topiclist[i].summary)");
      }
   ET(update idletasks);
   
   
   }
/* -------------------------------------------------------------- */
/* mark a topic as deleted */
ET_PROC( deleteq )
   {
   char sw[512];
   char *swp = sw;
   char *remoteadd, *remoteuser;
   unsigned long remoteip;   
   int confbutton;
   int i;
   int deltopic;

   /* find out which topic has been selected */
   swp = ET_STR(
      .listframe.list curselection
      );
      
   /* make sure there is something there */
   if (strcmp(swp, "") == 0)
      {
      ET(
         bell
         tk_dialog .viewdialog \
         "Error" \
         "You must select a topic" \
         "error" \
         0 \
         "Return"
      );
      return ET_OK;
      }
      
   deltopic = atoi(swp) + 1; /* master structure starts at 1 */
   
   /* check to see that this is one of our topics */
   
   remoteadd = getword(topiclist[deltopic].qid, 2);
   remoteip = inet_addr(remoteadd);
   remoteuser = getword(topiclist[deltopic].qid, 1);
   
   if (
      (remoteip != hostaddr) ||
      (strcmp(remoteuser, username) != 0)
      )
      {
      logger("user tried to delete question they did not create");
      ET(
         tk_dialog .deletedialog \
         "Question Not Deleted" \
         "You did not create that question, so you cannot delete it." \
         "error" \
         0 \
         "Dismiss"
      );
      free(remoteadd);
      free(remoteuser);
      return ET_OK;
      }
      
   free(remoteadd);
   free(remoteuser);

   /* make sure they really want to delete this */   
   confbutton = ET_INT( 
      tk_dialog .dialog1 \
         "Confirm Delete" \
         "Do you really want to delete the question: %q(topiclist[deltopic].summary)?" \
         "question" \
         0 \
         "Yes"  "No"
      );
   if (confbutton == 1)
      {
      return ET_OK;
      }
   
   /* send out the delete_question packet */
   numdeleted++;
   memcpy(&deltopics[numdeleted], &topiclist[deltopic], sizeof(struct topic));
   queue_question_delete(numdeleted);
   
   /* get rid of the local copy */
   /* shuffle all higher topics down one */
   for (i = deltopic; i < numtopics; i++)
      {
      memcpy(&topiclist[i], &topiclist[i+1], sizeof(struct topic));
      }
   /* reset the last space to zero */
   memset(&topiclist[numtopics], 0, sizeof(struct topic));
   numtopics--;
   
   /* update the cache file */
   writedata();
   
   /* update the display of topics */
   listtopics();
   
   
   return ET_OK;     
   }
  
/* -------------------------------------------------------------- */
/* view the details of the topic, and allow user to respond */

int quest(int thistopic)
   {
   int i, j, foundresponse;
   int k;
   int hasexpired;
   char *optiontext = NULL;
   time_t tmptime;
   time_t now, ntpnow;
   long int tdiff;

   /* if a question window exists, get rid of it */
   if (ET_INT(winfo exists .qmain) != 0)
      {
      ET(destroy .qmain);
      }
      
   /* if a results window exists, get rid of it */
   if (ET_INT(winfo exists .rmain) != 0)
      {
      ET(destroy .rmain);
      }
      
   if (ET_INT(winfo exists .textmain) != 0)
      {
      ET(destroy .textmain);
      }
      
   /* get the relevant data into the shared TCL variables */
   strcpy(current_summary, topiclist[thistopic].summary);
   strcpy(current_question, topiclist[thistopic].question);
   strcpy(current_author, topiclist[thistopic].email);
   strcpy(current_option, topiclist[thistopic].option);
   current_decay = topiclist[thistopic].decay_time;
   
   /* printf("Thistopic set to %d\n", thistopic); */
   ET(set tcltopic %d(thistopic));
   
   
   tmptime = ntptounix(topiclist[thistopic].time);
   strcpy(current_time , ctime(&tmptime));
   
   /* now display the question details */
   ET_INCLUDE( quest.tcl );
   
   /* clear out all the old button variables */
   for (i = 1; i <= MAX_RESPONSES; i++)
      {
      if (ET_INT(info exists button%d(i)))
         {
         ET(unset button%d(i));
         }
      }
      
   /* tell user how many responses are possible */
   
   if (strcmp(topiclist[thistopic].option, "text") == 0)
      {
      optiontext = strdup("Type Your Response");
      }
   else if (strcmp(topiclist[thistopic].option, "many") == 0)
      {
      optiontext = strdup("Select One or More Responses");
      }
   else
      {
      optiontext = strdup("Select One Response");
      }
   
   ET(
      message .qmain.optionframe.message \
      -text "%q(optiontext)" \
      -width "300" \
      -fg "green4"
      pack .qmain.optionframe.message -in .qmain.optionframe 
      );

   free(optiontext);
         
   foundresponse = 0; 
   hasexpired = 0;  
   for (j = 1; j <= nummyresponses; j++)
      {
      if (strcmp(myresponse[j].id, topiclist[thistopic].qid) == 0)
         {
         foundresponse = 1;
         break;
         /* j will be index to appropriate my response */
         }
      }
      
   /* match my response to global response list */
   if (foundresponse)
      {   
      /* lookup if this is expired or not */
      if (topiclist[thistopic].decay_time != 0)
         {
         time(&now);
         ntpnow = unixtontp(now);
         tdiff = ntpnow - myresponse[j].time;

         /* life time expressed in minutes so multiply by 60 */
         if (tdiff > (topiclist[thistopic].decay_time * DECAY_FACTOR))
            {
            hasexpired = 1;
            }
         }
      }
      
   
   /* now populate the display with the responses */
   
   if (strcmp(topiclist[thistopic].option, "text") == 0)
      {
      /* create an input box */
      ET(
         text .qmain.butframe.textwin \
            -width 40 \
            -height 10 \
            -wrap "word" 
         pack .qmain.butframe.textwin -side "top"
         );
         
      /* put old response in input box */
      if (foundresponse)
         {
         ET(
           .qmain.butframe.textwin insert end "%q(myresponse[j].response[1])\n"
            );
         }
      }
   else
      {
      for (i = 1; i <= MAX_RESPONSES; i++)
         {
         int butnum = 0;
         if (topiclist[thistopic].response[i] != NULL)
            {

            /* multiple responses */
            if (strcmp(topiclist[thistopic].option, "many") == 0)
               {
               butnum = i;

               if (foundresponse)
                  {
                  /* set the variables to the previous responses */
                  for (k = 1; k <= MAX_RESPONSES ; k++)
                     {
                     if ((myresponse[j].response[butnum] != NULL) && (topiclist[thistopic].response[k] != NULL))
                        {
                        if (strcmp(myresponse[j].response[butnum], topiclist[thistopic].response[k]) == 0)
                           {
                           ET(set button%d(k) "%d(k)");
                           break;
                           }
                        }
                     }
                  }


               ET(
                  checkbutton .qmain.butframe.button%d(i) \
                  -variable "button%d(butnum)" \
                  -onvalue "%d(i)" \
                  -text "%q(topiclist[thistopic].response[i])" \
                  -fg red
               );
              }


            /* single response */
            else
               {
               /* force the same button variable for all buttons */
               butnum = 1;

               if (foundresponse)
                  {
                  /* set the variable to the previous response */
                  for (k = 1; k <= MAX_RESPONSES ; k++)
                     {
                     if (strcmp(myresponse[j].response[1], 
                        topiclist[thistopic].response[k]) == 0)
                           {
                           break;
                           /* k will be right response number */
                           }
                     }
                  ET(set button%d(butnum) "%d(k)");
                  }

               ET(
                  radiobutton .qmain.butframe.button%d(i) \
                  -variable "button%d(butnum)" \
                  -value "%d(i)" \
                  -text "%q(topiclist[thistopic].response[i])" \
                  -fg red
               );
               }


            ET(
               pack .qmain.butframe.button%d(i) -side top -anchor w
            );
            }
        } /* button loop */
     }
     
   if (hasexpired)
      {
      ET(
         message .qmain.expireframe.expired \
            -text "Your response has expired, please submit again." \
            -fg "orange2" \
            -justify center \
            -width 600
         pack .qmain.expireframe.expired -in .qmain.expireframe -side top
      );
      }
   }
   
/* -------------------------------------------------------------- */
/* the interface to create a new question */

ET_PROC( createq )   
   {
   int typebutton;
   
   /* reset all the Tcl variables */
   strcpy(current_summary, "");
   strcpy(current_question, "");
   strcpy(current_author, "");
   strcpy(current_time, "");
   current_decay = 0;
   
   
   /* find out the type of question */
   
   typebutton = ET_INT(
      tk_dialog .qtype \
      "What type of question?" \
      "What type of question are you creating?" \
      "question" \
      0 \
      "Multiple Choice" "Text"
      );
   
   /* if multi choice */
   if (typebutton == 0)
      {
      ET_INCLUDE( create.tcl );
      }
   
   /* if text response */
   else
      {
      ET_INCLUDE( textcreate.tcl );
      }
   return ET_OK;
   }

/* -------------------------------------------------------------- */
/* add the new question to the data structure */

ET_PROC( addq )
   {
   char tmp[1024];
   char *fname = NULL;
   char *tmpp = tmp;
   int curtop, i, j;
   time_t now;
   struct passwd *pass, *getpwnam();
   char nametmp[TMPSTRLEN];
   char *resptmp = NULL;
   char *tmpquestion = NULL;
   
   /* check that we have enough in the data fields to proceed */
   if (strcmp(current_summary, "") == 0)
      {
      ET(
         tk_dialog .notopicdialog \
         "No Question Topic" \
         "You must enter a question topic." \
         "error" \
         0 \
         "Dismiss"
      );
      return ET_OK;
      }
      
   tmpquestion = 
     strdup(ET_STR(.cmain.entryframe.qframe.textwin get 1.0 end));
     
   if (strcmp(tmpquestion, "\n") == 0)
      {
      ET(
         tk_dialog .noquestiondialog \
         "No Question" \
         "You must enter a poll question." \
         "error" \
         0 \
         "Dismiss"
      );
      free(tmpquestion);
      return ET_OK;
      }
      
   numtopics++; /* adding a new topic */
   curtop = numtopics;
   
   /* get the data from TCL variables */
   
   topiclist[curtop].summary = strdup(current_summary);
   
   /* note: question can have embedded \n chars, and will always
            have \n at the end */
   topiclist[curtop].question = strdup(tmpquestion);
   
   free(tmpquestion);
     
   /* so remove all the \n characters */
   removenl(topiclist[curtop].question);
      
   topiclist[curtop].option = strdup( ET_STR(set type));
   topiclist[curtop].decay_time = current_decay;
   
   /* remind time stuff will be added later */
/*
 *    topiclist[curtop].remind_time = ET_INT(set remindtime);
 */
   topiclist[curtop].remind_time = 0;
   
   
   /* create the question ID */
   /* get the current time */
   time(&now);
   /* convert time to NTP format for storage */
   now = unixtontp(now);
   
   sprintf(tmpp, "%s %s %lu", username, hostip, now);
   topiclist[curtop].qid = strdup(tmpp);
   
   
   /*  perhaps should also read .RTPdefaults */
   /* get the email information from the OS */
#ifndef WIN32   
   pass = getpwnam(username);
   fname = strdup(pass->pw_gecos);
#else
   fname = strdup(username);
#endif
   
   sprintf(nametmp, "%s <%s@%s>", fname, username, hostname);
   free(fname);
   topiclist[curtop].email = strdup(nametmp);
   
   topiclist[curtop].time = now;
   
   /* use NTP time for ID and version */
   topiclist[curtop].version = now;
   
   topiclist[curtop].theme = strdup(theme);
   
   if (strcmp(topiclist[curtop].option, "text") != 0)
      {
      /* get the possible responses for the new question */
      j = 0;
      for (i = 1; i <= MAX_RESPONSES; i++)
         {
         resptmp = strdup( ET_STR(set resp%d(i)));
         if (strcmp(resptmp, "") != 0)
            {
            j++;
            topiclist[curtop].response[j] = strdup(resptmp);
            }
         free(resptmp);
         }
      }
      
   /* update the cache file */
   writedata();

   /* update the topic list */
   listtopics();

   /* and send the new question out to the world */   
   send_question_packet(curtop);
      
   /* get rid of the new question window */
   ET(destroy .cmain);
   
   return ET_OK;
   }


/* -------------------------------------------------------------- */
/* confirm that the user really wants to exit */

ET_PROC( confirm )   
   {
   int confbutton;
   
   confbutton = ET_INT( 
      tk_dialog .dialog1 \
         "Confirm Exit" \
         "Do you really want to quit?" \
         "question" \
         0 \
         "Yes"  "No" 
   	);

   
   if (confbutton == 0)
      {
      /* write the cache data file */
      writedata();
      exit(0); /* normal exit */
      }
   else
      {
      return ET_OK;
      }
   }

/* -------------------------------------------------------------- */
/* a place holder for functions that are missing */

ET_PROC( missing )   
   {
   ET(
      bell
      tk_dialog .missdialog \
      "Function is missing" \
      "Sorry, this function has not been built yet" \
      "error" \
      0 \
      "Return"
   );
   return ET_OK;
   }
   
   
/* -------------------------------------------------------------- */
/* show command line argument information */

usage( char *message)
   {
   printf("\n\aCommand line error: %s\n", message);
   printf("\nUsage: mpoll [-D] [-s] [-l] [-V] [-t ttl] [-C \"title\"] address/port\n\
    OPTIONS:\n\
      -D  : use default address and port\n\
      -s  : simulate results (useful for demos)\n\
      -V  : print version number and exit\n\
      -l  : print logging information (for debugging)\n\
      -C \"title\" :  the theme title for the questions\n\
      -t ttl : the multicast TTL (default = 15)\n\n");
   exit(2);
   }
   
/* -------------------------------------------------------------- */
/* learn local hostname and user information */

learn_local_info()
   {
   struct hostent *hstent;
#ifndef WIN32   
   uid_t uid;
   struct passwd* pswd;
   struct utsname name;
#else
#define _SYS_NMLN 9
   int uid;
   int nmsize;
   struct utsname {
      char sysname[_SYS_NMLN];
      char nodename[_SYS_NMLN];
      char release[_SYS_NMLN];
      char version[_SYS_NMLN];
      char machine[_SYS_NMLN];
   };
   struct utsname name;
#endif
   struct in_addr myaddr;


   /* find out information about the system */
   uname(&name);
   sprintf(systeminfo, "%s-%s-%s", name.sysname, name.release, name.machine);

   /*find our own address*/
   gethostname(hostname, BUFFER_SIZE);
   if (hostname[0] == '\0') 
      {
      fprintf(stderr, "gethostname failed!\n");
      exit(1);
      }
   hstent=(struct hostent*)gethostbyname(hostname);
   if (hstent == (struct hostent*) NULL) 
      {
      fprintf(stderr, "gethostbyname failed, hostname={%s}\n", hostname);
      exit(1);
      }
      
   /* store away our address in long it format */
   memcpy((char *)&hostaddr, (char *)hstent->h_addr_list[0], hstent->h_length);
   /* hostaddr is already in network byte order from use of gethost */
   
   /* and in IP dotted quad format */
   myaddr.s_addr = hostaddr;
   strcpy(hostip, (char *)inet_ntoa(myaddr));

   /* find the user's username */
   strcpy(username, "-"); /* default */
   
#ifndef WIN32   
   uid=getuid();
   pswd=getpwuid(uid);
   if(pswd != 0)
      {
      strcpy(username, pswd->pw_name);
      }
#else
   nmsize = sizeof(username);
   GetUserName(username, &nmsize);
   sprintf(messbuf, "Win32 username is %s\n", username);
   logger(messbuf);
#endif

   /* get the user's home directory */
   gethomedir();
   
      
	}
	

